@@ -23,6 +23,12 @@
 #include "file.h"
 #include "map.h"
 #include "scenario.h"
+#include "localize.h"
+
+#if DEBUG_CAMPAIGN
+#  include <stdio.h>
+#  include <stdlib.h>
+#endif
 
 /*
 ====================================================================
@@ -48,6 +54,8 @@ extern int weather_type_count;
 extern List *players;
 extern Font *log_font;
 extern Setup setup;
+extern int deploy_turn;
+extern List *unit_lib;
 
 /*
 ====================================================================
@@ -67,12 +75,14 @@ List *reinf = 0;        /* reinforcements -- units with a delay (if delay is 0
                            controls it. */
 Player *player = 0;     /* current player */
 int log_x, log_y;       /* position where to draw log info */
+char *scen_domain;	/* domain this scenario was loaded under */
 /* VICTORY CONDITIONS */
 char scen_result[64] = "";  /* the scenario result is saved here */
 char scen_message[128] = ""; /* the final scenario message is saved here */
 int vcond_check_type = 0;   /* test victory conditions this turn */
 VCond *vconds = 0;          /* victory conditions */
 int vcond_count = 0;
+static int *casualties;	/* sum of casualties grouped by unit class and player */
 
 /*
 ====================================================================
@@ -123,7 +133,7 @@ static int subcond_check( VSubCond *cond )
                is not fullfilled */
             list_reset( units );
             while ( ( unit = list_next( units ) ) )
-                if ( !player_is_ally( cond->player, unit->player ) )
+                if ( !unit->killed && !player_is_ally( cond->player, unit->player ) )
                     if ( unit->tag[0] != 0 && STRCMP( unit->tag, cond->tag ) )
                         return 0;
             return 1;
@@ -142,6 +152,45 @@ static int subcond_check( VSubCond *cond )
     return 0;
 }
 
+/** Set no_purchase for all nations that either have no purchasable
+ * units in current unit library or have neither flag nor unit under
+ * control (thus are not present in this scenario). */
+static void update_nations_purchase_flag()
+{
+	int i, x, y;
+	Unit_Lib_Entry *e = NULL;
+	Unit *u = NULL;
+	
+	/* clear flag -- misuse it for checks below */
+	for (i = 0; i < nation_count; i++)
+		nations[i].no_purchase = 0;
+	
+	/* mark all nations that have at least one entry assigned */
+	list_reset( unit_lib );
+	while ((e = list_next( unit_lib )))
+		if (e->nation != -1)
+			nations[e->nation].no_purchase |= 0x01;
+		
+	/* mark all nations that have at least one unit/flag under control */
+	list_reset( units );
+	while ((u = list_next( units )))
+		u->nation->no_purchase |= 0x02;
+	list_reset( reinf );
+	while ((u = list_next( reinf )))
+		u->nation->no_purchase |= 0x02;
+	for ( x = 0; x < map_w; x++ )
+		for ( y = 0; y < map_h; y++ )
+			if ( map[x][y].nation )
+				map[x][y].nation->no_purchase |= 0x02;
+			
+	/* map results - purchase okay if both flags set */
+	for (i = 0; i < nation_count; i++)
+		if (nations[i].no_purchase == 0x03)
+			nations[i].no_purchase = 0;
+		else 
+			nations[i].no_purchase = 1;
+}
+
 /*
 ====================================================================
 Publics
@@ -153,7 +202,7 @@ Publics
 Load a scenario.
 ====================================================================
 */
-int scen_load( char *fname )
+int scen_load( const char *fname )
 {
     char log_str[256];
     int unit_ref = 0;
@@ -163,28 +212,33 @@ int scen_load( char *fname )
     PData *pd_vcond, *pd_bind, *pd_vsubcond;
     List *entries, *values, *flags;
     char path[512];
-    char aux[128];
     char *str, *lib;
+    char *domain;
     Player *player = 0;
     Unit_Lib_Entry *unit_prop = 0, *trsp_prop = 0;
     Unit *unit;
     Unit unit_base; /* used to store values for unit */
+    int unit_delayed = 0;
+	
     scen_delete();
-    sprintf( path, "%s/scenarios/%s", SRC_DIR, fname );
+    sprintf( path, "%s/scenarios/%s", get_gamedir(), fname );
     if ( ( pd = parser_read_file( fname, path ) ) == 0 ) goto parser_failure;
+    free(scen_domain);
+    domain = scen_domain = determine_domain(pd, fname);
+    locale_load_domain(domain, 0/*FIXME*/);
     SDL_FillRect( sdl.screen, 0, 0x0 );
     log_font->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
     log_x = 2; log_y = 2;
-    sprintf( log_str, "*** Loading scenario '%s' ***", fname );
+    sprintf( log_str, tr("*** Loading scenario '%s' ***"), fname );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     /* get scenario info */
-    sprintf( log_str, "Loading Scenario Info");
+    sprintf( log_str, tr("Loading Scenario Info"));
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     scen_info = calloc( 1, sizeof( Scen_Info ) );
     scen_info->fname = strdup( fname );
-    if ( !parser_get_string( pd, "name", &scen_info->name ) ) goto parser_failure;
-    if ( !parser_get_string( pd, "desc", &scen_info->desc ) ) goto parser_failure;
-    if ( !parser_get_string( pd, "authors", &scen_info->authors ) ) goto parser_failure;
+    if ( !parser_get_localized_string( pd, "name", domain, &scen_info->name ) ) goto parser_failure;
+    if ( !parser_get_localized_string( pd, "desc", domain, &scen_info->desc ) ) goto parser_failure;
+    if ( !parser_get_localized_string( pd, "authors", domain, &scen_info->authors) ) goto parser_failure;
     if ( !parser_get_value( pd, "date", &str, 0 ) ) goto parser_failure;
     str_to_date( &scen_info->start_date, str );
     if ( !parser_get_int( pd, "turns", &scen_info->turn_limit ) ) goto parser_failure;
@@ -193,7 +247,7 @@ int scen_load( char *fname )
     if ( !parser_get_entries( pd, "players", &entries ) ) goto parser_failure;
     scen_info->player_count = entries->count;
     /* nations */
-    sprintf( log_str, "Loading Nations" );
+    sprintf( log_str, tr("Loading Nations") );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( !parser_get_value( pd, "nation_db", &str, 0 ) ) goto parser_failure;
     if ( !nations_load( str ) ) goto failure;
@@ -202,19 +256,19 @@ int scen_load( char *fname )
         /* check the scenario file itself but only for the main entry */
         sprintf(path, "../scenarios/%s", fname); 
         str = path;
-        sprintf( log_str, "Loading Main Unit Library '%s'", str );
+        sprintf( log_str, tr("Loading Main Unit Library '%s'"), str );
         write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
         if ( !unit_lib_load( str, UNIT_LIB_MAIN ) ) goto parser_failure;
     }
     else {
         if ( !parser_get_value( sub, "main", &str, 0 ) ) goto parser_failure;
-        sprintf( log_str, "Loading Main Unit Library '%s'", str );
+        sprintf( log_str, tr("Loading Main Unit Library '%s'"), str );
         write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
         if ( !unit_lib_load( str, UNIT_LIB_MAIN ) ) goto parser_failure;
         if ( parser_get_values( sub, "add", &values ) ) {
             list_reset( values );
             while ( ( lib = list_next( values ) ) ) {
-                sprintf( log_str, "Loading Additional Unit Library '%s'", lib );
+                sprintf( log_str, tr("Loading Additional Unit Library '%s'"), lib );
                 write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
                 if ( !unit_lib_load( lib, UNIT_LIB_ADD ) )
                     goto failure;
@@ -226,7 +280,7 @@ int scen_load( char *fname )
         sprintf(path, "../scenarios/%s", fname); /* check the scenario file itself */
         str = path;
     }
-    sprintf( log_str, "Loading Map '%s'", str );
+    sprintf( log_str, tr("Loading Map '%s'"), str );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( !map_load( str ) ) goto failure;
     weather = calloc( scen_info->turn_limit, sizeof( int ) );
@@ -243,21 +297,26 @@ int scen_load( char *fname )
         }
     }
     /* players */
-    sprintf( log_str, "Loading Players" );
+    sprintf( log_str, tr("Loading Players") );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( !parser_get_entries( pd, "players", &entries ) ) goto parser_failure;
     list_reset( entries );
     while ( ( sub = list_next( entries ) ) ) {
         /* create player */
         player = calloc( 1, sizeof( Player ) );
+	    
         player->id = strdup( sub->name );
-        if ( !parser_get_string( sub, "name", &player->name ) ) goto parser_failure;
+        if ( !parser_get_localized_string( sub, "name", domain, &player->name ) ) goto parser_failure;
         if ( !parser_get_values( sub, "nations", &values ) ) goto parser_failure;
         player->nation_count = values->count;
         player->nations = calloc( player->nation_count, sizeof( Nation* ) );
         list_reset( values );
         for ( i = 0; i < values->count; i++ )
             player->nations[i] = nation_find( list_next( values ) );
+	
+	/* unit limit (0 = no limit) */
+	parser_get_int(sub, "unit_limit", &player->unit_limit);
+	
         if ( !parser_get_value( sub, "orientation", &str, 0 ) ) goto parser_failure;
         if ( STRCMP( str, "right" ) ) /* alldirs not implemented yet */
             player->orient = UNIT_ORIENT_RIGHT;
@@ -281,6 +340,25 @@ int scen_load( char *fname )
             player->sea_trsp = unit_lib_find( str );
             if ( !parser_get_int( subsub, "count", &player->sea_trsp_count ) ) goto parser_failure;
         }
+	/* Load prestige info if any. If none is present show a warning. */ 
+	player->prestige_per_turn = calloc( scen_info->turn_limit, sizeof(int));
+	if (!player->prestige_per_turn) {
+		fprintf( stderr, tr("Out of memory\n") );
+		goto failure;
+	}
+	if ( parser_get_values( sub, "prestige", &values ) ) {
+		list_reset( values ); i = 0;
+		while ( ( str = list_next( values ) ) ) {
+			player->prestige_per_turn[i] = atoi(str);
+			i++;
+		}
+	} else
+		fprintf( stderr, tr("Oops: No prestige info for player %s "
+					"in scenario %s.\nMaybe you did not "
+					"convert scenarios after update?\n"), 
+					player->name, scen_info->name );
+	player->cur_prestige = 0; /* will be adjusted on turn begin */
+	    
         player_add( player ); player = 0;
     }
     /* set alliances */
@@ -296,7 +374,7 @@ int scen_load( char *fname )
         }
     }
     /* flags */
-    sprintf( log_str, "Loading Flags" );
+    sprintf( log_str, tr("Loading Flags") );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( !parser_get_entries( pd, "flags", &flags ) ) goto parser_failure;
     list_reset( flags );
@@ -314,7 +392,7 @@ int scen_load( char *fname )
     }
     /* victory conditions */
     scen_result[0] = 0;
-    sprintf( log_str, "Loading Victory Conditions" );
+    sprintf( log_str, tr("Loading Victory Conditions") );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     /* check type */
     vcond_check_type = VCOND_CHECK_EVERY_TURN;
@@ -343,7 +421,7 @@ int scen_load( char *fname )
             else
                 strcpy( vconds[i].result, "undefined" );
             if ( parser_get_value( pd_vcond, "message", &str, 0 ) )
-                strcpy_lt( vconds[i].message, str, 127 );
+                strcpy_lt( vconds[i].message, trd(domain, str), 127 );
             else
                 strcpy( vconds[i].message, "undefined" );
             /* and linkage */
@@ -454,7 +532,7 @@ int scen_load( char *fname )
             }
             /* no sub conditions at all? */
             if ( vconds[i].sub_and_count == 0 && vconds[i].sub_or_count == 0 ) {
-                fprintf( stderr, "No subconditions specified!\n" );
+                fprintf( stderr, tr("No subconditions specified!\n") );
                 goto failure;
             }
             /* next condition */
@@ -463,15 +541,15 @@ int scen_load( char *fname )
     }
     /* else condition (used if no other condition is fullfilled and scenario ends) */
     strcpy( vconds[0].result, "defeat" );
-    strcpy( vconds[0].message, "Defeat" );
+    strcpy( vconds[0].message, tr("Defeat") );
     if ( parser_get_pdata( pd, "result/else", &pd_vcond ) ) {
         if ( parser_get_value( pd_vcond, "result", &str, 0 ) )
             strcpy_lt( vconds[0].result, str, 63 );
         if ( parser_get_value( pd_vcond, "message", &str, 0 ) )
-            strcpy_lt( vconds[0].message, str, 127 );
+            strcpy_lt( vconds[0].message, trd(domain, str), 127 );
     }
     /* units */
-    sprintf( log_str, "Loading Units" );
+    sprintf( log_str, tr("Loading Units") );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     units = list_create( LIST_AUTO_DELETE, unit_delete );
     reinf = list_create( LIST_AUTO_DELETE, unit_delete );
@@ -483,35 +561,29 @@ int scen_load( char *fname )
         /* unit type */
         if ( !parser_get_value( sub, "id", &str, 0 ) ) goto parser_failure;
         if ( ( unit_prop = unit_lib_find( str ) ) == 0 ) {
-            fprintf( stderr, "%s: unit entry not found\n", str );
+            fprintf( stderr, tr("%s: unit entry not found\n"), str );
             goto failure;
         }
         memset( &unit_base, 0, sizeof( Unit ) );
         /* nation & player */
         if ( !parser_get_value( sub, "nation", &str, 0 ) ) goto parser_failure;
         if ( ( unit_base.nation = nation_find( str ) ) == 0 ) {
-            fprintf( stderr, "%s: not a nation\n", str );
+            fprintf( stderr, tr("%s: not a nation\n"), str );
             goto failure;
         }
         if ( ( unit_base.player = player_get_by_nation( unit_base.nation ) ) == 0 ) {
-            fprintf( stderr, "%s: no player controls this nation\n", unit_base.nation->name );
+            fprintf( stderr, tr("%s: no player controls this nation\n"), unit_base.nation->name );
             goto failure;
         }
         /* name */
-        switch ( ( unit_ref + 1 ) % 10 ) {
-            case 1: sprintf( aux, "%ist %s", unit_ref + 1, unit_prop->name ); break;
-            case 2: sprintf( aux, "%ind %s", unit_ref + 1, unit_prop->name ); break;
-            case 3: sprintf( aux, "%ird %s", unit_ref + 1, unit_prop->name ); break;
-            default: sprintf( aux, "%ith %s", unit_ref + 1, unit_prop->name ); break;
-        }
-        strcpy_lt( unit_base.name, aux, 20 );
+        unit_set_generic_name( &unit_base, unit_ref + 1, unit_prop->name );
         /* delay */
-        parser_get_int( sub, "delay", &unit_base.delay );
+        unit_delayed = parser_get_int( sub, "delay", &unit_base.delay );
         /* position */
-        if ( !parser_get_int( sub, "x", &unit_base.x ) && !unit_base.delay ) goto parser_failure;
-        if ( !parser_get_int( sub, "y", &unit_base.y ) && !unit_base.delay ) goto parser_failure;
-        if ( !unit_base.delay && ( unit_base.x <= 0 || unit_base.y <= 0 || unit_base.x >= map_w - 1 || unit_base.y >= map_h - 1 ) ) {
-            fprintf( stderr, "%s: out of map: ignored\n", unit_base.name );
+        if ( !parser_get_int( sub, "x", &unit_base.x ) && !unit_delayed ) goto parser_failure;
+        if ( !parser_get_int( sub, "y", &unit_base.y ) && !unit_delayed ) goto parser_failure;
+        if ( !unit_delayed && ( unit_base.x <= 0 || unit_base.y <= 0 || unit_base.x >= map_w - 1 || unit_base.y >= map_h - 1 ) ) {
+            fprintf( stderr, tr("%s: out of map: ignored\n"), unit_base.name );
             continue;  
         }
         /* strengt, entrenchment, experience */
@@ -544,12 +616,12 @@ int scen_load( char *fname )
         /* actual unit */
         unit = unit_create( unit_prop, trsp_prop, &unit_base );
         /* put unit to active or reinforcements list */
-        if ( unit->delay == 0 ) {
+        if ( !unit_delayed ) {
             list_add( units, unit );
             /* add unit to map */
             map_insert_unit( unit );
         }
-        else 
+        else if (!config.purchase) /* no fixed reinfs with purchase enabled */
             list_add( reinf, unit );
         /* adjust transporter count */
         if ( unit->embark == EMBARK_SEA ) {
@@ -563,7 +635,39 @@ int scen_load( char *fname )
             }
         unit_ref++;
     }
+    /* load deployment hexes */
+    if ( parser_get_entries( pd, "deployfields", &entries ) ) {
+        list_reset( entries );
+        while ( ( sub = list_next( entries ) ) ) {
+            Player *pl;
+            int plidx;
+            if ( !parser_get_value( sub, "id", &str, 0 ) ) goto parser_failure;
+            pl = player_get_by_id( str );
+            if ( !pl ) goto parser_failure;
+            plidx = player_get_index( pl );
+            if ( parser_get_values( sub, "coordinates", &values ) && values->count>0 )
+            {
+                list_reset( values );
+                while ( ( lib = list_next( values ) ) ) {
+                    if (!strcmp(lib,"default"))
+                        { x=-1; y=-1; }
+                    else if (!strcmp(lib,"none"))
+                        { pl->no_init_deploy = 1; continue; }
+                    else
+                        get_coord( lib, &x, &y );
+                    map_set_deploy_field( x, y, plidx );
+                }
+            }
+        }
+    }
     parser_free( &pd );
+
+    casualties = calloc( scen_info->player_count * unit_class_count, sizeof casualties[0] );
+    deploy_turn = config.deploy_turn;
+    
+    /* check which nations may do purchase for this scenario */
+    update_nations_purchase_flag();
+    
     return 1;
 parser_failure:        
     fprintf( stderr, "%s\n", parser_get_error() );
@@ -580,27 +684,32 @@ failure:
 Load a scenario description.
 ====================================================================
 */
-char* scen_load_info( char *fname )
+char* scen_load_info( const char *fname )
 {
     PData *pd, *sub;
     char path[512];
     List *entries;
     char *name, *desc, *turns, count[4], *str;
     char *info = 0;
+    char *domain = 0;
     int i;
-    sprintf( path, "%s/scenarios/%s", SRC_DIR, fname );
+    sprintf( path, "%s/scenarios/%s", get_gamedir(), fname );
     if ( ( pd = parser_read_file( fname, path ) ) == 0 ) goto parser_failure;
     /* title and description */
+    domain = determine_domain(pd, fname);
+    locale_load_domain(domain, 0/*FIXME*/);
     if ( !parser_get_value( pd, "name", &name, 0 ) ) goto parser_failure;
+    name = (char *)trd(domain, name);
     if ( !parser_get_value( pd, "desc", &desc, 0 ) ) goto parser_failure;
+    desc = (char *)trd(domain, desc);
     if ( !parser_get_entries( pd, "players", &entries ) ) goto parser_failure;
     if ( !parser_get_value( pd, "turns", &turns, 0 ) ) goto parser_failure;
     sprintf( count, "%i",  entries->count );
     if ( ( info = calloc( strlen( name ) + strlen( desc ) + strlen( count ) + strlen( turns ) + 30, sizeof( char ) ) ) == 0 ) {
-        fprintf( stderr, "Out of memory\n" );
+        fprintf( stderr, tr("Out of memory\n") );
         goto failure;
     }
-    sprintf( info, "%s##%s##%s Turns#%s Players", name, desc, turns, count );
+    sprintf( info, tr("%s##%s##%s Turns#%s Players"), name, desc, turns, count );
     /* set setup */
     scen_clear_setup();
     strcpy( setup.fname, fname );
@@ -611,7 +720,8 @@ char* scen_load_info( char *fname )
     /* load the player ctrls */
     list_reset( entries ); i = 0;
     while ( ( sub = list_next( entries ) ) ) {
-        if ( !parser_get_string( sub, "name", &setup.names[i] ) ) goto parser_failure;
+        if ( !parser_get_value( sub, "name", &str, 0 ) ) goto parser_failure;
+        setup.names[i] = strdup(trd(domain, str));
         if ( !parser_get_value( sub, "control", &str, 0 ) ) goto parser_failure;
         if ( STRCMP( str, "cpu" ) )
             setup.ctrl[i] = PLAYER_CTRL_CPU;
@@ -622,12 +732,14 @@ char* scen_load_info( char *fname )
         i++;
     }
     parser_free( &pd );
+    free(domain);
     return info;
 parser_failure:        
     fprintf( stderr, "%s\n", parser_get_error() );
 failure:
     if ( pd ) parser_free( &pd );
     if ( info ) free( info );
+    free(domain);
     return 0;
 }
 
@@ -688,6 +800,17 @@ Delete scenario
 void scen_delete()
 {
     int i;
+#if 0
+    int j;
+    printf( "Casualties:\n" );
+    for (i = 0; casualties && i < scen_info->player_count; i++) {
+        printf( "%s Side\n", player_get_by_index(i)->name );
+        for (j = 0; j < unit_class_count; j++)
+            printf( "%d\t%s\n", scen_get_casualties( i, j ), unit_classes[j].name );
+    }
+#endif
+    free( casualties );
+    casualties = 0;
     turn = 0;
     cur_weather = 0;
     if ( scen_info ) {
@@ -730,12 +853,16 @@ void scen_delete()
     unit_lib_delete();
     players_delete();
     map_delete();
+
+    free(scen_domain);
+    scen_domain = 0;
 }
 
 /*
 ====================================================================
-Check if unit is destroyed, set the current attack and movement.
-If SCEN_PREP_UNIT_FIRST is passed the entrenchment is not modified.
+Set the current attack and movement. If SCEN_PREP_UNIT_FIRST is 
+passed the entrenchment is not modified. Units are neither supplied
+or killed here!
 ====================================================================
 */
 void scen_prep_unit( Unit *unit, int type )
@@ -743,12 +870,6 @@ void scen_prep_unit( Unit *unit, int type )
     int min_entr, max_entr;
     /* unit may not be undeployed */
     unit->fresh_deploy = 0;
-    /* crash air planes */
-    if ( unit_check_fuel_usage( unit ) && unit->cur_fuel == 0 && unit->sel_prop->flags & FLYING ) 
-        unit->remove = 1;
-    /* set supply level */
-    if ( unit->unused && unit->supply_level > 0  )
-        unit_supply( unit, UNIT_SUPPLY_ALL );
     /* remove turn suppression */
     unit->turn_suppr = 0;
     /* allow actions */
@@ -756,8 +877,11 @@ void scen_prep_unit( Unit *unit, int type )
     unit->cur_mov = unit->sel_prop->mov;
     if ( unit_check_ground_trsp( unit ) )
         unit->cur_mov = unit->trsp_prop.mov;
-    /* if we have bad weather the relation between mov : fuel is 1 : 2 */
-    if ( weather_types[scen_get_weather()].flags & DOUBLE_FUEL_COST ) {
+    /* if we have bad weather the relation between mov : fuel is 1 : 2 
+     * for ground units
+     */
+    if ( !(unit->sel_prop->flags & SWIMMING) && !(unit->sel_prop->flags & FLYING)
+         && (weather_types[scen_get_weather()].flags & DOUBLE_FUEL_COST) ) {
         if ( unit_check_fuel_usage( unit ) && unit->cur_fuel < unit->cur_mov * 2 )
             unit->cur_mov = unit->cur_fuel / 2;
     }
@@ -773,9 +897,7 @@ void scen_prep_unit( Unit *unit, int type )
         if ( unit->entr < min_entr )
             unit->entr = min_entr;
         else
-            if ( unit->entr >= max_entr )
-                unit->entr = max_entr;
-            else
+            if ( unit->entr < max_entr )
                 unit->entr++;
     }
 }
@@ -793,6 +915,22 @@ first one).
 int scen_check_result( int after_last_turn )
 {
     int i, j, and_okay, or_okay;
+#if DEBUG_CAMPAIGN
+    char fname[512];
+    FILE *f;
+    snprintf(fname, sizeof fname, "%s/.lgames/.scenresult", getenv("HOME"));
+    f = fopen(fname, "r");
+    if (f) {
+        unsigned len;
+        scen_result[0] = '\0';
+        fgets(scen_result, sizeof scen_result, f);
+        fclose(f);
+        len = strlen(scen_result);
+        if (len > 0 && scen_result[len-1] == '\n') scen_result[len-1] = '\0';
+        strcpy(scen_message, scen_result);
+        if (len > 0) return 1;
+    }
+#endif
     if ( vcond_check_type == VCOND_CHECK_EVERY_TURN || after_last_turn ) {
         for ( i = 1; i < vcond_count; i++ ) {
             /* AND binding */
@@ -869,79 +1007,9 @@ Check the supply level of a unit. (hex tiles with SUPPLY_GROUND
 have 100% supply)
 ====================================================================
 */
-int is_close_to_harbor( Unit *unit )
-{
-    int i;
-    if ( map_is_allied_depot( &map[unit->x][unit->y], unit ) )
-        return 1;
-    for ( i = 0; i < 6; i++ )
-        if ( map_is_allied_depot( map_get_close_hex( unit->x, unit->y, i ), unit ) )
-            return 1;
-    return 0;
-}
-int is_close_to_airfield( Unit *unit )
-{
-    int i;
-    if ( map_is_allied_depot( &map[unit->x][unit->y], unit ) )
-        return 1;
-    for ( i = 0; i < 6; i++ )
-        if ( map_is_allied_depot( map_get_close_hex( unit->x, unit->y, i ), unit ) )
-            return 1;
-    return 0;
-}
 void scen_adjust_unit_supply_level( Unit *unit )
 {
-    int x, y, w, h, i, j;
-    int flag_supply_level;
-    /* flying and swimming units get a 100% supply if near an airfield or a harbour */
-    if ( unit->sel_prop->flags & SWIMMING ) {
-        unit->supply_level = 0;
-        if ( is_close_to_harbor( unit ) ) unit->supply_level = 100;
-    }
-    else
-    if ( unit->sel_prop->flags & FLYING ) {
-        unit->supply_level = 0;
-        if ( is_close_to_airfield( unit ) ) unit->supply_level = 100;
-    }
-    else {
-        /* ground units get a 100% close to a flag and looses about 10% for each title it gets away */
-        /* test all flags within a region x-10,y-10,20,20 about their distance */
-        /* get region first */
-        x = unit->x - 10; y = unit->y - 10; w = 20; h = 20;
-        if ( x < 0 ) { w += x; x = 0; }
-        if ( y < 0 ) { h += y; y = 0; }
-        if ( x + w > map_w ) w = map_w - x;
-        if ( y + h > map_h ) h = map_h - y;
-        /* now check flags */
-        unit->supply_level = 0;
-        for ( i = x; i < x + w; i++ )
-            for ( j = y; j < y + h; j++ )
-                if ( map[i][j].player && player_is_ally( unit->player, map[i][j].player ) ) {
-                    flag_supply_level = get_dist( unit->x, unit->y, i, j );
-                    if ( flag_supply_level < 2 ) flag_supply_level = 100;
-                    else {
-                        flag_supply_level = 100 - ( flag_supply_level - 1 ) * 10;
-                        if ( flag_supply_level < 0 ) flag_supply_level = 0;
-                    }
-                    if ( flag_supply_level > unit->supply_level )
-                        unit->supply_level = flag_supply_level;
-                }
-    }
-    /* if hostile influence is 1 supply's cut in half if influence >1 no supply possible */
-    if ( unit->sel_prop->flags & FLYING ) {
-        if ( mask[ unit->x][ unit->y ].air_infl > 1 || mask[ unit->x][ unit->y ].infl > 1 )
-            unit->supply_level = 0;
-        else
-            if ( mask[ unit->x][unit->y ].air_infl == 1 || mask[ unit->x][ unit->y ].infl == 1 )
-                unit->supply_level /= 2;
-    }
-    else {
-        if ( mask[ unit->x][ unit->y ].infl == 1 )
-            unit->supply_level /= 2;
-        else
-            if ( mask[ unit->x][ unit->y ].infl > 1 )
-                unit->supply_level = 0;
-    }
+    unit->supply_level = map_get_unit_supply_level( unit->x, unit->y, unit );
 }
 
 /*
@@ -988,3 +1056,35 @@ void scen_get_date( char *date_str )
     }
 }
 
+/*
+====================================================================
+Get/Add casualties for unit class and player.
+====================================================================
+*/
+int scen_get_casualties( int player, int class )
+{
+    if (!casualties || player < 0 || player >= scen_info->player_count
+        || class < 0 || class >= unit_class_count)
+        return 0;
+    return casualties[player*unit_class_count + class];
+}
+int scen_inc_casualties( int player, int class )
+{
+    if (!casualties || player < 0 || player >= scen_info->player_count
+         || class < 0 || class >= unit_class_count)
+        return 0;
+    return ++casualties[player*unit_class_count + class];
+}
+/*
+====================================================================
+Add casualties for unit. Regard unit and transport classes.
+====================================================================
+*/
+int scen_inc_casualties_for_unit( Unit *unit )
+{
+    int player = player_get_index( unit->player );
+    int cnt = scen_inc_casualties( player, unit->prop.class );
+    if ( unit->trsp_prop.id )
+        cnt = scen_inc_casualties( player, unit->trsp_prop.class );
+    return cnt;
+}
