@@ -15,10 +15,14 @@
  *                                                                         *
  ***************************************************************************/
 
-#include <SDL.h>
+#include "sdl.h"
+
+#include "misc.h"
+
+#include <assert.h>
 #include <stdlib.h>
 #include <string.h>
-#include "sdl.h"
+#include "localize.h"
 
 extern int  term_game;
 
@@ -31,15 +35,15 @@ int cur_time, last_time;
 /* sdl surface */
 
 /* return full path of bitmap */
-inline void get_full_bmp_path( char *full_path, char *file_name )
+inline void get_full_bmp_path( char *full_path, const char *file_name )
 {
-    sprintf(full_path,  "%s/gfx/%s", SRC_DIR, file_name );
+    sprintf(full_path,  "%s/gfx/%s", get_gamedir(), file_name );
 }
 
 /*
     load a surface from file putting it in soft or hardware mem
 */
-SDL_Surface* load_surf(char *fname, int f)
+SDL_Surface* load_surf(const char *fname, int f)
 {
     SDL_Surface *buf;
     SDL_Surface *new_sur;
@@ -190,120 +194,209 @@ Uint32 get_pixel( SDL_Surface *surf, int x, int y )
 /* sdl font */
 
 /* return full font path */
-void get_full_font_path( char *path, char *file_name )
+void get_full_font_path( char *path, const char *file_name )
 {
     strcpy( path, file_name );
 /*    sprintf(path, "./gfx/fonts/%s", file_name ); */
 }
 
 /*
-    load a font using the width values in the file
-*/
-Font* load_font(char *fname)
+ * Loads glyphs into a font from file 'fname'.
+ *
+ * Font format description:
+ *
+ * Each font file is a pixmap comprised of a single row of subsequent glyphs
+ * defining the appearence of the associated character code.
+ *
+ * The height of the font is implicitly defined by the height of the pixmap.
+ *
+ * Each glyph is defined by a pixmap as depicted in the following
+ * example (where each letter represents the respective rgb-value):
+ *
+ * (1) -----> ........
+ *            ........
+ *            ..####..
+ *            .#....#.
+ *            .#....#.
+ *            .#....#.
+ *            .#....#.
+ *            .#....#.
+ *            .#.#..#.
+ *            .#..#.#.
+ *            ..####..
+ *            ......#.
+ * (2, 3) --> S.......
+ *
+ * (1): The top left pixel defines the transparency pixel. All pixels of
+ * the same color as (1) are treated as transparent when the glyph is
+ * rendered.
+ * Note that (1) will only be checked for the *first* glyph in
+ * the file and be applied to all other glyphs.
+ *
+ * (2): The bottom left pixel defines the starting column of a particular
+ * glyph within the font pixmap. By scanning these pixels, the font loader
+ * is able to determine the width of the glyph.
+ * The color is either #FF00FF for a valid glyph, or #FF0000 for an invalid
+ * glyph (in this case it will not be rendered,
+ * regardless of what it contains otherwise).
+ *
+ * (3): For the very first glyph, (2) has a special meaning. It provides
+ * some basic information about the font to be loaded.
+ * The r-value specifies the starting code of the first glyph.
+ * The g- and b-values are reserved and must be (0)
+ *
+ * The glyph-to-character-code-mapping is done by starting with the code
+ * specified by (3). Then the bottom line is scanned: For every occurrence of
+ * (2) the code will be incremented, and appropriate offset and width values
+ * be calculated. The total count of characters will be determined by the
+ * count of glyphs contained within the file.
+ *
+ * For display, (2) and (3) will be assumed to have color (1).
+ */
+void font_load_glyphs(Font *font, const char *fname)
 {
-    Font    *fnt = 0;
-    FILE    *file = 0;
+#if SDL_BYTEORDER == SDL_BIG_ENDIAN
+#  define R_INDEX 1
+#  define G_INDEX 2
+#  define B_INDEX 3
+#else
+#  define R_INDEX 2
+#  define G_INDEX 1
+#  define B_INDEX 0
+#endif
+#define AT_RGB(pixel, idx) (((Uint8 *)&(pixel))+(idx))
     char    path[512];
     int     i;
+    Uint32  transparency_key;
+    Uint8   start_code, reserved1, reserved2;
+    SDL_Surface *new_glyphs;
 
     get_full_font_path( path, fname );
 
-    fnt = malloc(sizeof(Font));
-    if (fnt == 0) {
-        fprintf(stderr, "ERR: sfnt_load: not enough memory\n");
+    if ((new_glyphs = load_surf(path, SDL_SWSURFACE)) == 0) {
+        fprintf(stderr, tr("Cannot load new glyphs surface: %s\n"), SDL_GetError());
+        exit(1);
+    }
+    /* use (1) as transparency key */
+    transparency_key = get_pixel( new_glyphs, 0, 0 );
+
+    /* evaluate (3) */
+    SDL_GetRGB(get_pixel( new_glyphs, 0, new_glyphs->h - 1 ), new_glyphs->format, &start_code, &reserved1, &reserved2);
+    // bail out early and consistently for invalid input to minimise the amount
+    // of font files out of spec
+    if (reserved1 || reserved2) {
+        fprintf(stderr, tr("Invalid font file: %d, %d\n"), reserved1, reserved2);
         exit(1);
     }
 
-    if ((fnt->pic = load_surf(path, SDL_HWSURFACE)) == 0)
-        exit(1);
-    /* use very first pixel as transparency key */
-    SDL_SetColorKey( fnt->pic, SDL_SRCCOLORKEY, get_pixel( fnt->pic, 0, 0 ) );		
-		
-    fnt->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
-    fnt->color = 0x00FFFFFF;
-    fnt->height = fnt->pic->h;
+    /* update font data */
+    /* FIXME: don't blindly override, merge smartly */
+    font->height = new_glyphs->h;
 	
-    /* table */
-    file = fopen(path, "r");
-    fseek(file, -1, SEEK_END);
-    fread(&fnt->offset, 1, 1, file);
-#ifdef SDL_DEBUG
-    printf("offset: %i\n", fnt->offset);
-#endif
-    fseek(file, -2, SEEK_END);
-    fread(&fnt->length, 1, 1, file);
-#ifdef SDL_DEBUG
-    printf("number: %i\n", fnt->length);
-#endif
-    fseek(file, -2 - fnt->length, SEEK_END);
-    fread(fnt->char_width, 1, fnt->length, file);
-#ifdef SDL_DEBUG
-    printf("letter width: %i\n", fnt->length);
-    for (i = 0; i < fnt->length; i++)
-        printf("%i ", fnt->char_width[i]);
-    printf("\n");
-#endif
-    fclose(file);
+    /* cut prevalent glyph row at where to insert the new glyph row,
+     * and insert it, thus overriding only the glyph range that is defined
+     * by new_glyphs.
+     */
+    {
+        /* total width of previous conserved glyphs */
+        int pre_width = font->char_offset[start_code];
+        /* total width of following conserved glyphs */
+        int post_width = (font->pic ? font->pic->w : 0) - pre_width;
+        unsigned code = start_code;
+        int fixup;	/* amount of pixels following offsets have to be fixed up */
+        SDL_Surface *dest;
+        
+        /* override widths and offsets of new glyphs */
+        /* concurrently calculate width of following conserved glyphs */
+        for (i = 0; i < new_glyphs->w; i++) {
+            Uint32 pixel = 0;
+            int valid_glyph;
+            SDL_GetRGB(get_pixel(new_glyphs, i, new_glyphs->h - 1), new_glyphs->format, AT_RGB(pixel, R_INDEX), AT_RGB(pixel, G_INDEX), AT_RGB(pixel, B_INDEX));
+            pixel &= 0xf8f8f8;	/* cope with RGB565 and other perversions */
+            if (i != 0 && pixel != 0xf800f8 && pixel != 0xf80000) continue;
 
-    /* letter offsets */
-    fnt->char_offset[0] = 0;
-    for (i = 1; i < fnt->length; i++)	
-        fnt->char_offset[i] = fnt->char_offset[i - 1] + fnt->char_width[i - 1];
-	
-    /* allowed keys */
-    memset(fnt->keys, 0, 256);
-    for (i = 0; i < fnt->length; i++) {
-        fnt->keys[i + fnt->offset] = 1;
+            if (code > 256) {
+                fprintf(stderr, tr("font '%s' contains too many glyphs\n"), path);
+                break;
+            }
+
+            valid_glyph = i == 0 || pixel != 0xf80000;
+            set_pixel(new_glyphs, i, new_glyphs->h - 1, transparency_key);
+            
+            font->char_offset[code] = pre_width + i;
+            font->keys[code] = valid_glyph;
+            code++;
+            
+        }
+        
+        fixup = pre_width + i - font->char_offset[code];
+        post_width -= font->char_offset[code] - font->char_offset[start_code];
+        
+        /* now the gory part:
+         * 1. create a new surface large enough to hold conserved and new glyphs.
+         * 2. blit the conserved previous part.
+         * 3. blit the conserved following part.
+         * 4. blit the new glyphs.
+         */
+        dest = create_surf(pre_width + new_glyphs->w + post_width,
+        		   font->pic ? font->pic->h : new_glyphs->h, SDL_HWSURFACE);
+        if (!dest) {
+            fprintf(stderr, tr("could not create font surface: %s\n"), SDL_GetError());
+            exit(1);
+        }
+
+        if (pre_width > 0) {
+            assert(font->pic);
+            DEST(dest, 0, 0, pre_width, font->height);
+            SOURCE(font->pic, 0, 0);
+            blit_surf();
+        }
+        
+        if (post_width > 0) {
+            assert(font->pic);
+            DEST(dest, font->char_offset[code] + fixup, 0, post_width, font->height);
+            SOURCE(font->pic, font->char_offset[code], 0);
+            blit_surf();
+        }
+        
+        DEST(dest, font->char_offset[start_code], 0, new_glyphs->w, font->height);
+        FULL_SOURCE(new_glyphs);
+        blit_surf();
+        
+        /* replace old row with newly composed row */
+        SDL_FreeSurface(font->pic);
+        font->pic = dest;
+
+        /* fix up offsets of successors */
+        for (i = code; i < 256; i++)
+            font->char_offset[code] += fixup;
+        font->width += fixup;
     }
-	
-    fnt->last_x = fnt->last_y = fnt->last_width = fnt->last_height = 0;
-    return fnt;
+    
+    SDL_SetColorKey( font->pic, SDL_SRCCOLORKEY, transparency_key );
+
+    SDL_FreeSurface(new_glyphs);
+#undef R_INDEX
+#undef G_INDEX
+#undef B_INDEX
+#undef AT_RGB
 }
 
-/*
-    load a font with fixed size
-*/
-Font* load_fixed_font(char *f, int off, int len, int w)
+/**
+ * create a new font from font file 'fname'
+ */
+Font* load_font(const char *fname)
 {
-    int     i;
-    Font    *fnt;
-    char    path[512];
-
-    get_full_font_path( path, f );
-
-    fnt = malloc(sizeof(Font));
+    Font *fnt = calloc(1, sizeof(Font));
     if (fnt == 0) {
-        fprintf(stderr, "ERR: sfnt_load: not enough memory\n");
+        fprintf(stderr, tr("ERR: %s: not enough memory\n"), __FUNCTION__);
         exit(1);
     }
-
-    if ((fnt->pic = load_surf(path, SDL_HWSURFACE)) == 0)
-        exit(1);
-    /* use very first pixel as transparency key */
-    SDL_SetColorKey( fnt->pic, SDL_SRCCOLORKEY, get_pixel( fnt->pic, 0, 0 ) );		
-
+    
     fnt->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
     fnt->color = 0x00FFFFFF;
-    fnt->height = fnt->pic->h;
-	
-	fnt->offset = off;
-	fnt->length = len;
-	
-	for (i = 0; i < len; i++)
-	    fnt->char_width[i] = w;
-	
-    /* letter offsets */
-    fnt->char_offset[0] = 0;
-    for (i = 1; i < fnt->length; i++)	
-        fnt->char_offset[i] = fnt->char_offset[i - 1] + w;
-	
-    /* allowed keys*/
-    memset(fnt->keys, 0, 256);
-    for (i = 0; i < fnt->length; i++) {
-        fnt->keys[i + fnt->offset] = 1;
-    }
-	
-    fnt->last_x = fnt->last_y = fnt->last_width = fnt->last_height = 0;
+    
+    font_load_glyphs(fnt, fname);
     return fnt;
 }
 
@@ -322,49 +415,47 @@ void free_font(Font **fnt)
 /*
     write something with transparency
 */
-int write_text(Font *fnt, SDL_Surface *dest, int x, int y, char *str, int alpha)
+int write_text(Font *fnt, SDL_Surface *dest, int x, int y, const char *str, int alpha)
 {
-    int	c_abs;
     int len = strlen(str);
-    int pix_len = 0;
+    int pix_len = text_width(fnt, str);
     int px = x, py = y;
     int i;
     SDL_Surface *spf = SDL_GetVideoSurface();
+    const int * const ofs = fnt->char_offset;
 	
-    pix_len = text_width(fnt, str);
-	for (i = 0; i < len; i++)
-	    if (!fnt->keys[(int)str[i]])
-	        str[i] = ' ';
-
     /* alignment */
     if (fnt->align & ALIGN_X_CENTER)
         px -= pix_len >> 1;
-    else
-        if (fnt->align & ALIGN_X_RIGHT)
-            px -= pix_len;
+    else if (fnt->align & ALIGN_X_RIGHT)
+        px -= pix_len;
     if (fnt->align & ALIGN_Y_CENTER)
         py -= (fnt->height >> 1 ) + 1;
     else
         if (fnt->align & ALIGN_Y_BOTTOM)
             py -= fnt->height;
 
-    fnt->last_x = px; if (fnt->last_x < 0) fnt->last_x = 0;
-    fnt->last_y = py; if (fnt->last_y < 0) fnt->last_y = 0;
-    fnt->last_width = pix_len; if (fnt->last_x + fnt->last_width >= spf->w) fnt->last_width = spf->w - fnt->last_x;
-    fnt->last_height = fnt->height; if (fnt->last_y + fnt->last_height >= spf->h) fnt->last_height = spf->h - fnt->last_y;
+    fnt->last_x = MAXIMUM(px, 0);
+    fnt->last_y = MAXIMUM(py, 0);
+    fnt->last_width = MINIMUM(pix_len, spf->w - fnt->last_x);
+    fnt->last_height = MINIMUM(fnt->height, spf->h - fnt->last_y);
 
     if (alpha != 0)
         SDL_SetAlpha(fnt->pic, SDL_SRCALPHA, alpha);
     else
         SDL_SetAlpha(fnt->pic, 0, 0);
     for (i = 0; i < len; i++) {
-       	c_abs = str[i] - fnt->offset;
-       	DEST(dest, px, py, fnt->char_width[c_abs], fnt->height);
-       	SOURCE(fnt->pic, fnt->char_offset[c_abs], 0);
-       	blit_surf();
-        px += fnt->char_width[c_abs];
+        unsigned c = (unsigned char)str[i];
+        if (!fnt->keys[c]) c = '\177';
+        {
+            const int w = ofs[c+1] - ofs[c];
+            DEST(dest, px, py, w, fnt->height);
+            SOURCE(fnt->pic, ofs[c], 0);
+            blit_surf();
+            px += w;
+        }
     }
-	
+    
     return 0;
 }
 
@@ -374,7 +465,7 @@ Write string to x, y and modify y so that it draws to the
 next line.
 ====================================================================
 */
-void write_line( SDL_Surface *surf, Font *font, char *str, int x, int *y )
+void write_line( SDL_Surface *surf, Font *font, const char *str, int x, int *y )
 {
     write_text( font, surf, x, *y, str, 255 );
     *y += font->height;
@@ -407,23 +498,24 @@ SDL_Rect last_write_rect(Font *fnt)
     return rect;
 }
 
+inline int  char_width(Font *fnt, char c)
+{
+    unsigned i = (unsigned char)c;
+    return fnt->char_offset[i + 1] - fnt->char_offset[i];
+}
+
 /*
     return the text width in pixels
 */
-int text_width(Font *fnt, char *str)
+int text_width(Font *fnt, const char *str)
 {
     unsigned int i;
     int pix_len = 0;
-    for (i = 0; i < strlen(str); i++)
-        pix_len += fnt->char_width[str[i] - fnt->offset];
+    for (i = strlen(str); i > 0; )
+        pix_len += char_width(fnt, str[--i]);
     return pix_len;
 }
 
-int  char_width(Font *fnt, char c)
-{
-    return fnt->char_width[c - fnt->offset];
-}
-
 /* sdl */
 
 /*
@@ -443,7 +535,7 @@ void init_sdl( int f )
         exit(1);
     }
     SDL_EnableUNICODE(1);
-    atexit(SDL_Quit);
+    atexit(quit_sdl);
     /* create empty cursor */
     empty_cursor = create_cursor( 16, 16, 8, 8,
                                   "                "
@@ -472,6 +564,50 @@ void quit_sdl()
 {
     if (sdl.screen) SDL_FreeSurface(sdl.screen);
     if ( empty_cursor ) SDL_FreeCursor( empty_cursor );
+    if (sdl.vmodes) free(sdl.vmodes);
+    SDL_Quit();
+    printf("SDL finalized\n");
+}
+
+/** Get list of all video modes. Allocate @vmi and return number of 
+ * entries. */
+int get_video_modes( VideoModeInfo **vmi )
+{
+	SDL_Rect **modes = SDL_ListModes(NULL, SDL_FULLSCREEN);
+	int i, nmodes = 0;
+	int bpp = SDL_GetVideoInfo()->vfmt->BitsPerPixel;
+	
+	*vmi = NULL;
+	
+	if (modes == NULL || modes == (SDL_Rect **)-1) {
+		VideoModeInfo stdvmi[2] = {
+			{ 640, 480, 32, 0 },
+			{ 640, 480, 32, 1 }
+		};
+		fprintf(stderr,"No video modes available or specified.\n");
+		/* offer at least 640x480 */
+		nmodes = 2;
+		*vmi = calloc(nmodes,sizeof(VideoModeInfo));
+		(*vmi)[0] = stdvmi[0];
+		(*vmi)[1] = stdvmi[1];
+		return nmodes;
+	}
+	
+	/* count number of modes */
+	for (nmodes = 0; modes[nmodes]; nmodes++);
+	nmodes *= 2; /* each as window and fullscreen */
+	
+	*vmi = calloc(nmodes,sizeof(VideoModeInfo));
+	for (i = 0; i < nmodes/2; i++) {
+		VideoModeInfo *info = &((*vmi)[i*2]);
+		info->width = modes[i]->w;
+		info->height = modes[i]->h;
+		info->depth = bpp;
+		info->fullscreen = 0;
+		(*vmi)[i*2+1] = *info;
+		(*vmi)[i*2+1].fullscreen = 1;
+	}
+	return nmodes;
 }
 
 /*
@@ -497,7 +633,7 @@ int set_video_mode( int width, int height, int fullscreen )
     /* set video mode */
     if ( fullscreen ) flags |= SDL_FULLSCREEN;
     if ( ( depth = SDL_VideoModeOK( width, height, depth, flags ) ) == 0 ) {
-        fprintf( stderr, "Requested mode %ix%i, Fullscreen: %i unavailable\n",
+        fprintf( stderr, tr("Requested mode %ix%i, Fullscreen: %i unavailable\n"),
                  width, height, fullscreen );
         sdl.screen = SDL_SetVideoMode( 640, 480, 16, SDL_SWSURFACE );
     }
@@ -730,9 +866,9 @@ inline void flip_screen()
 /* cursor */
 
 /* creates cursor */
-SDL_Cursor* create_cursor( int width, int height, int hot_x, int hot_y, char *source )
+SDL_Cursor* create_cursor( int width, int height, int hot_x, int hot_y, const char *source )
 {
-    char *mask = 0, *data = 0;
+    unsigned char *mask = 0, *data = 0;
     SDL_Cursor *cursor = 0;
     int i, j, k;
     char data_byte, mask_byte;
