@@ -17,9 +17,11 @@
  
 #include <dirent.h>
 #include "lgeneral.h"
+#include "windows.h"
+#include "unit.h"
+#include "purchase_dlg.h"
 #include "gui.h"
 #include "nation.h"
-#include "unit.h"
 #include "player.h"
 #include "date.h"
 #include "map.h"
@@ -229,8 +231,10 @@ void strat_map_delete()
     free_surf( &unit_layer );
     free_surf( &flag_layer );
     free_surf( &blink_dot );
-    if ( dots )
+    if ( dots ) {
         list_delete( dots );
+        dots = 0;
+    }
 }
 
 /*
