@@ -16,17 +16,21 @@
  ***************************************************************************/
 
 #include "lgeneral.h"
+#include "list.h"
 #include "unit.h"
+#include "localize.h"
 
 /*
 ====================================================================
 Externals
 ====================================================================
 */
+extern int scen_get_weather( void );
 extern Sdl sdl;
 extern Config config;
 extern int trgt_type_count;
 extern List *vis_units; /* units known zo current player */
+extern List *units;
 extern int cur_weather;
 extern Weather_Type *weather_types;
 extern Unit *cur_target;
@@ -88,10 +92,10 @@ static int unit_apply_damage( Unit *unit, int damage, int suppr, Unit *attacker
         unit->str = 0;
         return 0;
     }
-    if ( attacker && attacker->sel_prop->flags & BOMBER ) {
+    if ( attacker && attacker->sel_prop->flags & TURN_SUPPR ) {
         unit->turn_suppr += suppr;
-        if ( unit->str - unit->turn_suppr < 0 ) {
-            unit->turn_suppr = unit->str;
+        if ( unit->str - unit->turn_suppr - unit->suppr < 0 ) {
+            unit->turn_suppr = unit->str - unit->suppr;
             return 0;
         }
     }
@@ -105,6 +109,54 @@ static int unit_apply_damage( Unit *unit, int damage, int suppr, Unit *attacker
     return unit_get_cur_str( unit );
 }
 
+/** Add prestige for player of unit @unit, based on properties of attacked 
+ * unit @target and the damage @target_dam inflicted to it. */
+static void unit_gain_prestige( Unit *unit, const Unit *target, int target_dam )
+{
+	int gain = 0;
+	
+	/* I played around with PG a little but did not figure out how the
+	 * formula works. Some points seem to hold true though:
+	 * - own damage/unit loss does not give prestige penalty
+	 *   (therefore we don't consider unit damage here)
+	 * - if target is damaged but not destroyed then some random prestige 
+	 *   is gained based on cost, experience and damage
+	 * - if target is destroyed then more prestige is gained based on
+	 *   cost and experience with just a little random modification
+	 *
+	 * FIXME: the formula used when unit is destroyed seems to be quite
+	 * good, but the one for mere damage not so much especially when
+	 * experience increases: rolling the dice more times does not increase
+	 * chance very much for low cost units so we give less prestige for 
+	 * cheap, experienced units compared to PG... */ 
+
+	if (target->str == 0) {
+		/* if unit destroyed: ((1/24 of cost) * exp) +- 10% */
+		gain = ((target->exp_level + 1) * target->prop.cost * 10 + 
+							120/*round up*/) / 240;
+		gain = ((111 - DICE(21)) * gain) / 100;
+		if (gain == 0)
+			gain = 1;
+	} else {
+		/* if damaged: try half damage multiplied by experience throws
+		 * on 50-dice against one tenth of cost. at maximum four times 
+		 * experience can be gained */
+		int throws = ((target_dam + 1) / 2) * (target->exp_level+1);
+		int i, limit = 4 * (target->exp_level+1) - 1;
+		for (i = 0; i < throws; i++)
+			if( DICE(50) < target->prop.cost/10 ) {
+				gain++;
+				if (gain == limit)
+					break;
+			}
+	}
+	
+	unit->player->cur_prestige += gain;
+	//printf("%s attacked %s (%d damage%s): prestige +%d\n",
+	//			unit->name, target->name, target_dam,
+	//			(target->str==0)?", killed":"", gain);
+}
+
 /*
 ====================================================================
 Execute a single fight (no defensive fire check) with random
@@ -120,6 +172,7 @@ enum {
 };
 static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_rugged )
 {
+    int unit_old_str = unit->str;//, target_old_str = target->str;
     int unit_old_ini = unit->sel_prop->ini, target_old_ini = target->sel_prop->ini;
     int unit_dam = 0, unit_suppr = 0, target_dam = 0, target_suppr = 0;
     int rugged_chance, rugged_def = 0;
@@ -130,38 +183,45 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
     if ( real && type == UNIT_ACTIVE_ATTACK )
         if ( unit_check_rugged_def( unit, target ) || ( force_rugged && unit_is_close( unit, target ) ) ) {
             rugged_chance = unit_get_rugged_def_chance( unit, target );
-            if ( RANDOM( 1, 100 ) <= rugged_chance || force_rugged )
+            if ( DICE(100) <= rugged_chance || force_rugged )
                 rugged_def = 1;
         }
     /* PG's formula for initiative is
        min { base initiative, terrain max initiative } +
        ( exp_level + 1 ) / 2 + D3 */
-    if ( !(unit->sel_prop->flags & FLYING) )
-        unit->sel_prop->ini = MINIMUM( unit->sel_prop->ini, unit->terrain->max_ini );
-    unit->sel_prop->ini += ( unit->exp_level + 1  ) / 2;
-    if ( !(target->sel_prop->flags & FLYING) )
+    /* against aircrafts the initiative is used since terrain does not matter */
+    /* target's terrain is used for fight */
+    if ( !(unit->sel_prop->flags & FLYING) && !(target->sel_prop->flags & FLYING) )
+    {
+        unit->sel_prop->ini = MINIMUM( unit->sel_prop->ini, target->terrain->max_ini );
         target->sel_prop->ini = MINIMUM( target->sel_prop->ini, target->terrain->max_ini );
+    }
+    unit->sel_prop->ini += ( unit->exp_level + 1  ) / 2;
     target->sel_prop->ini += ( target->exp_level + 1  ) / 2;
     /* special initiative rules:
        antitank inits attack tank|recon: atk 0, def 99
+       tank inits attack against anti-tank: atk 0, def 99
        defensive fire: atk 99, def 0
        submarine attacks: atk 99, def 0
        ranged attack: atk 99, def 0
        rugged defense: atk 0
-       air unit attacks air defense: atk = def */
+       air unit attacks air defense: atk = def 
+       non-art vs art: atk 0, def 99 */
     if ( unit->sel_prop->flags & ANTI_TANK )
         if ( target->sel_prop->flags & TANK ) {
             unit->sel_prop->ini = 0;
             target->sel_prop->ini = 99;
         }
-    if ( unit->sel_prop->flags & ( DIVING | ARTILLERY | AIR_DEFENSE ) || 
+    if ( (unit->sel_prop->flags&DIVING) || 
+         (unit->sel_prop->flags&ARTILLERY) || 
+         (unit->sel_prop->flags&AIR_DEFENSE) || 
          type == UNIT_DEFENSIVE_ATTACK
     ) {
         unit->sel_prop->ini = 99;
         target->sel_prop->ini = 0;
     }
     if ( unit->sel_prop->flags & FLYING )
-        if ( target->sel_prop->flags & AIR_DEFENSE ) 
+        if ( target->sel_prop->flags & AIR_DEFENSE )
             unit->sel_prop->ini = target->sel_prop->ini;
     if ( rugged_def )
         unit->sel_prop->ini = 0;
@@ -169,8 +229,8 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
         target->sel_prop->ini = 99;
     /* the dice is rolled after these changes */
     if ( real ) {
-        unit->sel_prop->ini += RANDOM( 1, 3 );
-        target->sel_prop->ini += RANDOM( 1, 3 );
+        unit->sel_prop->ini += DICE(3);
+        target->sel_prop->ini += DICE(3);
     }
 #ifdef DEBUG_ATTACK
     if ( real ) {
@@ -184,13 +244,18 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
 #endif
     /* in a real combat a submarine may evade */
     if ( real && type == UNIT_ACTIVE_ATTACK && ( target->sel_prop->flags & DIVING ) ) { 
-        if ( RANDOM( 1, 10 ) <= 7 + ( target->exp_level - unit->exp_level ) / 2 )
+        if ( DICE(10) <= 7 + ( target->exp_level - unit->exp_level ) / 2 )
+        {
             strike = ATK_NO_STRIKE;
+            ret |= AR_EVADED;
+        }
         else
             strike = ATK_UNIT_FIRST;
+#ifdef DEBUG_ATTACK
         printf ( "\nSubmarine Evasion: %s (%i%%)\n", 
                  (strike==ATK_NO_STRIKE)?"yes":"no",
                  10 * (7 + ( target->exp_level - unit->exp_level ) / 2) );
+#endif
     }
     else
     /* who is first? */
@@ -204,19 +269,19 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
     /* the one with the highest initiative begins first if not defensive fire or artillery */
     if ( strike == ATK_BOTH_STRIKE ) {
         /* both strike at the same time */
-        unit_get_damage( unit, target, type, real, rugged_def, &target_dam, &target_suppr );
+        unit_get_damage( unit, unit, target, type, real, rugged_def, &target_dam, &target_suppr );
         if ( unit_check_attack( target, unit, UNIT_PASSIVE_ATTACK ) )
-            unit_get_damage( target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
+            unit_get_damage( unit, target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
         unit_apply_damage( target, target_dam, target_suppr, unit );
         unit_apply_damage( unit, unit_dam, unit_suppr, target );
     }
     else
         if ( strike == ATK_UNIT_FIRST ) {
             /* unit strikes first */
-            unit_get_damage( unit, target, type, real, rugged_def, &target_dam, &target_suppr );
+            unit_get_damage( unit, unit, target, type, real, rugged_def, &target_dam, &target_suppr );
             if ( unit_apply_damage( target, target_dam, target_suppr, unit ) )
                 if ( unit_check_attack( target, unit, UNIT_PASSIVE_ATTACK ) && type != UNIT_DEFENSIVE_ATTACK ) {
-                    unit_get_damage( target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
+                    unit_get_damage( unit, target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
                     unit_apply_damage( unit, unit_dam, unit_suppr, target );
                 }
         }
@@ -224,12 +289,12 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
             if ( strike == ATK_TARGET_FIRST ) {
                 /* target strikes first */
                 if ( unit_check_attack( target, unit, UNIT_PASSIVE_ATTACK ) ) {
-                    unit_get_damage( target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
+                    unit_get_damage( unit, target, unit, UNIT_PASSIVE_ATTACK, real, rugged_def, &unit_dam, &unit_suppr );
                     if ( !unit_apply_damage( unit, unit_dam, unit_suppr, target ) )
                         ret |= AR_UNIT_ATTACK_BROKEN_UP;
                 }
                 if ( unit_get_cur_str( unit ) > 0 ) {
-                    unit_get_damage( unit, target, type, real, rugged_def, &target_dam, &target_suppr );
+                    unit_get_damage( unit, unit, target, type, real, rugged_def, &target_dam, &target_suppr );
                     unit_apply_damage( target, target_dam, target_suppr, unit );
                 }
             }
@@ -249,14 +314,21 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
     if ( real ) {
         /* cost ammo */
         if ( config.supply ) {
-            unit->cur_ammo--;
+            //if (DICE(10)<=target_old_str)
+                unit->cur_ammo--;
             if ( unit_check_attack( target, unit, UNIT_PASSIVE_ATTACK ) && target->cur_ammo > 0 )
-                target->cur_ammo--;
+                //if (DICE(10)<=unit_old_str)
+                    target->cur_ammo--;
         }
         /* costs attack */
         if ( unit->cur_atk_count > 0 ) unit->cur_atk_count--;
-        /* loose entrenchment */
-        if ( target->entr > 0 ) target->entr--;
+        /* target: loose entrenchment if damage was taken or with a unit->str*10% chance */
+        if ( target->entr > 0 ) 
+            if (target_dam > 0 || DICE(10)<=unit_old_str)
+                target->entr--;
+        /* attacker looses entrenchment if it got hurt */
+        if ( unit->entr > 0 && unit_dam > 0 )
+            unit->entr--;
         /* gain experience */
         exp_mod = target->exp_level - unit->exp_level;
         if ( exp_mod < 1 ) exp_mod = 1;
@@ -268,6 +340,9 @@ static int unit_attack( Unit *unit, Unit *target, int type, int real, int force_
             unit_add_exp( unit, 10 );
             unit_add_exp( target, 10 );
         }
+	/* gain prestige */
+	unit_gain_prestige( unit, target, target_dam );
+	unit_gain_prestige( target, unit, unit_dam );
         /* adjust life bars */
         update_bar( unit );
         update_bar( target );
@@ -351,6 +426,19 @@ void unit_delete( void *ptr )
     free( unit );
 }
 
+/*
+====================================================================
+Give unit a generic name.
+====================================================================
+*/
+void unit_set_generic_name( Unit *unit, int number, const char *stem )
+{
+    char numbuf[8];
+    
+    locale_write_ordinal_number(numbuf, sizeof numbuf, number);
+    snprintf(unit->name, 24, "%s %s", numbuf, stem);
+}
+
 /*
 ====================================================================
 Update unit icon according to it's orientation.
@@ -401,7 +489,7 @@ int unit_check_supply( Unit *unit, int type, int *missing_ammo, int *missing_fue
         *missing_ammo = 0;
     if ( missing_fuel )
         *missing_fuel = 0;
-    /* no supply near all already moved? */
+    /* no supply near or already moved? */
     if ( unit->embark == EMBARK_SEA || unit->embark == EMBARK_AIR ) return 0;
     if ( unit->supply_level == 0 ) return 0;
     if ( !unit->unused ) return 0;
@@ -428,19 +516,22 @@ int unit_check_supply( Unit *unit, int type, int *missing_ammo, int *missing_fue
 
 /*
 ====================================================================
-Supply percentage of maximum fuel/ammo/both
+Supply percentage of maximum fuel/ammo/both.
+_intern does not block movement etc.
 Return True if unit was supplied.
 ====================================================================
 */
-int unit_supply( Unit *unit, int type )
+int unit_supply_intern( Unit *unit, int type )
 {
-    int amount_ammo, amount_fuel, max;
+    int amount_ammo, amount_fuel, max, supply_amount;
     int supplied = 0;
     /* ammo */
     if ( type == UNIT_SUPPLY_AMMO || type == UNIT_SUPPLY_ALL )
     if ( unit_check_supply( unit, UNIT_SUPPLY_AMMO, &amount_ammo, &amount_fuel ) ) {
         max = unit->cur_ammo + amount_ammo ;
-        unit->cur_ammo += unit->supply_level * max / 100;
+        supply_amount = unit->supply_level * max / 100;
+        if ( supply_amount == 0 ) supply_amount = 1; /* at least one */
+        unit->cur_ammo += supply_amount;
         if ( unit->cur_ammo > max ) unit->cur_ammo = max;
         supplied = 1;
     }
@@ -448,11 +539,18 @@ int unit_supply( Unit *unit, int type )
     if ( type == UNIT_SUPPLY_FUEL || type == UNIT_SUPPLY_ALL )
     if ( unit_check_supply( unit, UNIT_SUPPLY_FUEL, &amount_ammo, &amount_fuel ) ) {
         max = unit->cur_fuel + amount_fuel ;
-        unit->cur_fuel += unit->supply_level * max / 100;
+        supply_amount = unit->supply_level * max / 100;
+        if ( supply_amount == 0 ) supply_amount = 1; /* at least one */
+        unit->cur_fuel += supply_amount;
         if ( unit->cur_fuel > max ) unit->cur_fuel = max;
         supplied = 1;
     }
-    if ( supplied ) {
+    return supplied;
+}
+int unit_supply( Unit *unit, int type )
+{
+    int supplied = unit_supply_intern(unit,type);
+    if (supplied) {
         /* no other actions allowed */
         unit->unused = 0; unit->cur_mov = 0; unit->cur_atk_count = 0;
     }
@@ -577,6 +675,7 @@ int unit_check_attack( Unit *unit, Unit *target, int type )
     }
     else {
         /* counter-attack */
+        if ( unit->cur_ammo <= 0 ) return 0;
         if ( !unit_is_close( unit, target ) && get_dist( unit->x, unit->y, target->x, target->y ) > unit->sel_prop->rng ) return 0;
         if ( unit->sel_prop->atks[target->sel_prop->trgt_type] == 0 ) return 0;
         /* artillery may only defend against close units */
@@ -595,14 +694,19 @@ int unit_check_attack( Unit *unit, Unit *target, int type )
 ====================================================================
 Compute damage/supression the target takes when unit attacks
 the target. No properties will be changed. If 'real' is set
-the dices are rolled else it's a stochastical prediction.
+the dices are rolled else it's a stochastical prediction. 
+'aggressor' is the unit that initiated the attack, either 'unit'
+or 'target'. It is not always 'unit' as 'unit' and 'target are 
+switched for get_damage depending on whether there is a striking
+back and who had the highest initiative.
 ====================================================================
 */
-void unit_get_damage( Unit *unit, Unit *target, int type, 
+void unit_get_damage( Unit *aggressor, Unit *unit, Unit *target, 
+                      int type, 
                       int real, int rugged_def,
                       int *damage, int *suppr )
 {
-    int atk_strength, max_roll;
+    int atk_strength, max_roll, min_roll, die_mod;
     int atk_grade, def_grade, diff, result;
     float suppr_chance, kill_chance;
     /* use PG's formula to compute the attack/defense grade*/
@@ -638,12 +742,19 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
         def_grade = target->sel_prop->def_air;
     else {
         def_grade = target->sel_prop->def_grnd;
+        /* apply close defense? */
         if ( unit->sel_prop->flags & INFANTRY )
             if ( !( target->sel_prop->flags & INFANTRY ) )
                 if ( !( target->sel_prop->flags & FLYING ) )
                     if ( !( target->sel_prop->flags & SWIMMING ) )
-                        if ( unit->terrain->flags[cur_weather] & INF_CLOSE_DEF )
+                    {
+                        if ( target == aggressor )
+                        if ( unit->terrain->flags[cur_weather]&INF_CLOSE_DEF )
                             def_grade = target->sel_prop->def_cls;
+                        if ( unit == aggressor )
+                        if ( target->terrain->flags[cur_weather]&INF_CLOSE_DEF )
+                            def_grade = target->sel_prop->def_cls;
+                    }
     }
 #ifdef DEBUG_ATTACK
     if ( real ) printf( "  base:   %2i\n", def_grade );
@@ -651,15 +762,24 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
 #endif
     /* experience */
     def_grade += target->exp_level;
-    /* attacker on a river? */
+    /* attacker on a river or swamp? */
     if ( !(unit->sel_prop->flags & FLYING) )
     if ( !(unit->sel_prop->flags & SWIMMING) )
     if ( !(target->sel_prop->flags & FLYING) )
-    if ( unit->terrain->flags[cur_weather] & RIVER ) {
-        def_grade += 4;
+    {
+        if ( unit->terrain->flags[cur_weather] & SWAMP ) 
+        {
+            def_grade += 4;
 #ifdef DEBUG_ATTACK
-        if ( real ) printf( "  river:  +4\n" );
+            if ( real ) printf( "  swamp:  +4\n" );
 #endif
+        } else
+        if ( unit->terrain->flags[cur_weather] & RIVER ) {
+            def_grade += 4;
+#ifdef DEBUG_ATTACK
+            if ( real ) printf( "  river:  +4\n" );
+#endif
+        }
     }
     /* rugged defense? */
     if ( type == UNIT_ACTIVE_ATTACK && rugged_def ) {
@@ -703,7 +823,7 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
         if ( unit->sel_prop->flags & ARTILLERY ) {
             def_grade += 3;
 #ifdef DEBUG_ATTACK
-            if ( real ) printf( "  artdef: +3\n" );
+            if ( real ) printf( "  def vs art: +3\n" );
 #endif
         }
     /* infantry versus anti_tank? */
@@ -711,22 +831,19 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
         if ( unit->sel_prop->flags & ANTI_TANK ) {
             def_grade += 2;
 #ifdef DEBUG_ATTACK
-            if ( real ) printf( "  antitnk:+3\n" );
+            if ( real ) printf( "  antitnk:+2\n" );
 #endif
         }
+    /* no fuel makes attacker less effective */
+    if ( unit_check_fuel_usage( unit ) && unit->cur_fuel == 0 )
+    {
+        def_grade += 4;
+#ifdef DEBUG_ATTACK
+            if ( real ) printf( "  lowfuel:+4\n" );
+#endif
+    }
     /* attacker strength */
     atk_strength = unit_get_cur_str( unit );
-    if ( weather_types[cur_weather].flags & CUT_STRENGTH )
-        atk_strength /= 2;
-    else
-    if ( unit_check_fuel_usage( unit ) && unit->cur_fuel == 0 )
-        atk_strength /= 2;
-    else {
-        if ( type == UNIT_DEFENSIVE_ATTACK )
-            if ( !(unit->sel_prop->flags & INTERCEPTOR) )
-                if ( !(unit->sel_prop->flags & ARTILLERY) || unit->sel_prop->rng > 1 )
-                    atk_strength /= 2;
-    }
 #ifdef DEBUG_ATTACK
     if ( real && atk_strength != unit_get_cur_str( unit ) )
         printf( "---\n%s with half strength\n", unit->name );
@@ -735,12 +852,12 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
         get difference between attack and defense
         strike for each strength point with 
           if ( diff <= 4 ) 
-              D20 + 4
+              D20 + diff
           else
               D20 + 4 + 0.4 * ( diff - 4 )
         suppr_fire flag set: 1-10 miss, 11-18 suppr, 19+ kill
         normal: 1-10 miss, 11-12 suppr, 13+ kill */
-    diff = atk_grade - def_grade;
+    diff = atk_grade - def_grade; if ( diff < -7 ) diff = -7;
     *damage = 0; *suppr = 0;
 #ifdef DEBUG_ATTACK
     if ( real ) {
@@ -750,26 +867,24 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
 #endif
     /* get the chances for suppression and kills (computed here
        to use also for debug info */
-    max_roll = 20 + ( diff <= 4 ? diff : 4 + 2 * ( diff - 4 ) / 5 );
+    suppr_chance = kill_chance = 0;
+    die_mod = ( diff <= 4 ? diff : 4 + 2 * ( diff - 4 ) / 5 );
+    min_roll = 1 + die_mod; max_roll = 20 + die_mod;
     /* get chances for suppression and kills */
     if ( unit->sel_prop->flags & SUPPR_FIRE ) {
-        if ( max_roll >= 18 )
-            suppr_chance = 8.0 / max_roll;
-        else
-            suppr_chance = 1 - 10.0 / max_roll;
-        if ( suppr_chance < 0 ) suppr_chance = 0;
-        kill_chance  = 1 - 18.0 / max_roll;
-        if ( kill_chance < 0 ) kill_chance = 0;
+        int limit = (type==UNIT_DEFENSIVE_ATTACK)?20:18;
+        if (limit-min_roll>=0)
+            suppr_chance = 0.05*(MINIMUM(limit,max_roll)-MAXIMUM(11,min_roll)+1);
+        if (max_roll>limit)
+            kill_chance = 0.05*(max_roll-MAXIMUM(limit+1,min_roll)+1);
     }
     else {
-        if ( max_roll >= 12 )
-            suppr_chance = 2.0 / max_roll;
-        else
-            suppr_chance = 1 - 10.0 / max_roll;
-        if ( suppr_chance < 0 ) suppr_chance = 0;
-        kill_chance  = 1 - 12.0 / max_roll;
-        if ( kill_chance < 0 ) kill_chance = 0;
+        if (12-min_roll>=0)
+            suppr_chance = 0.05*(MINIMUM(12,max_roll)-MAXIMUM(11,min_roll)+1);
+        if (max_roll>12)
+            kill_chance = 0.05*(max_roll-MAXIMUM(13,min_roll)+1);
     }
+    if (suppr_chance<0) suppr_chance=0; if (kill_chance<0) kill_chance=0;
     if ( real ) {
 #ifdef DEBUG_ATTACK
         printf( "Roll: D20 + %i (Kill: %i%%, Suppr: %i%%)\n", 
@@ -778,14 +893,15 @@ void unit_get_damage( Unit *unit, Unit *target, int type,
 #endif
         while ( atk_strength-- > 0 ) {
             if ( diff <= 4 )
-                result = RANDOM( 1, 20 ) + diff;
+                result = DICE(20) + diff;
             else
-                result = RANDOM( 1, 20 ) + 4 + 2 * ( diff - 4 ) / 5;
+                result = DICE(20) + 4 + 2 * ( diff - 4 ) / 5;
             if ( unit->sel_prop->flags & SUPPR_FIRE ) {
-                if ( result >= 11 && result <= 18 )
+                int limit = (type==UNIT_DEFENSIVE_ATTACK)?20:18;
+                if ( result >= 11 && result <= limit )
                     (*suppr)++;
                 else
-                    if ( result >= 19 )
+                    if ( result >= limit+1 )
                         (*damage)++;
             }
             else {
@@ -846,7 +962,7 @@ void unit_get_expected_losses( Unit *unit, Unit *target, int *unit_damage, int *
     /* let defensive fire go to work (no chance to defend against this) */
     list_reset( df_units );
     while ( ( df = list_next( df_units ) ) ) {
-        unit_get_damage( df, unit, UNIT_DEFENSIVE_ATTACK, 0, 0, &damage, &suppr );
+        unit_get_damage( unit, df, unit, UNIT_DEFENSIVE_ATTACK, 0, 0, &damage, &suppr );
         if ( !unit_apply_damage( unit, damage, suppr, 0 ) ) break;
     }
     /* actual fight if attack has strength remaining */
@@ -867,6 +983,11 @@ This function checks 'units' for supporters of 'target'
 that will give defensive fire to before the real battle
 'unit' vs 'target' takes place. These units are put to 'df_units'
 (which is not created here)
+
+Only adjacent units are defended and only ONE (randomly chosen) 
+supporter will actually fire as long as it does not have the flag
+'full_def_str'. (This means, all units with that flag plus one 
+normal supporter will fire.)
 ====================================================================
 */
 void unit_get_df_units( Unit *unit, Unit *target, List *units, List *df_units )
@@ -876,6 +997,7 @@ void unit_get_df_units( Unit *unit, Unit *target, List *units, List *df_units )
     if ( unit->sel_prop->flags & FLYING ) {
         list_reset( units );
         while ( ( entry = list_next( units ) ) ) {
+            if ( entry->killed ) continue;
             if ( entry == target ) continue;
             if ( entry == unit ) continue;
             /* bombers -- intercepting impossibly covered by unit_check_attack() */
@@ -889,20 +1011,27 @@ void unit_get_df_units( Unit *unit, Unit *target, List *units, List *df_units )
                             }
             /* air-defense */
             if ( entry->sel_prop->flags & AIR_DEFENSE )
-                if ( unit_check_attack( entry, unit, UNIT_DEFENSIVE_ATTACK ) )
-                    list_add( df_units, entry );
+                /* FlaK will not give support when an air-to-air attack is
+                 * taking place. First, in reality it would prove distastrous,
+                 * second, Panzer General doesn't allow it, either.
+                 */
+                if ( !(target->sel_prop->flags & FLYING) )
+                    if ( unit_is_close( target, entry ) ) /* adjacent help only */
+                        if ( unit_check_attack( entry, unit, UNIT_DEFENSIVE_ATTACK ) )
+                            list_add( df_units, entry );
         }
     }
-    else  {
-        /* artillery for ground units */
+    else if ( unit->sel_prop->rng==0 ) { 
+        /* artillery for melee combat; if unit attacks ranged, there is no 
+           support */
         list_reset( units );
         while ( ( entry = list_next( units ) ) ) {
+            if ( entry->killed ) continue;
             if ( entry == target ) continue;
             if ( entry == unit ) continue;
-            /* sturmgeschuetze (assault tanks?) will support all friendly
-               units standing beside them although the target may be 
-               out of range: and this is done with FULL strength in 
-               opposite to other artillery */
+            /* HACK: An artillery with range 1 cannot support adjacent units but
+               should do so. So we allow to give defensive fire on a range of 2
+               like a normal artillery */
             if ( entry->sel_prop->flags & ARTILLERY && entry->sel_prop->rng == 1 )
                 if ( unit_is_close( target, entry ) )
                     if ( player_is_ally( entry->player, target->player ) )
@@ -912,10 +1041,18 @@ void unit_get_df_units( Unit *unit, Unit *target, List *units, List *df_units )
                         }
             /* normal artillery */
             if ( entry->sel_prop->flags & ARTILLERY )
-                if ( unit_check_attack( entry, unit, UNIT_DEFENSIVE_ATTACK ) )
-                    list_add( df_units, entry );
+                if ( unit_is_close( target, entry ) ) /* adjacent help only */
+                    if ( unit_check_attack( entry, unit, UNIT_DEFENSIVE_ATTACK ) )
+                        list_add( df_units, entry );
         }
     }
+    /* randomly remove all support but one */
+    if (df_units->count>0)
+    {
+        entry = list_get(df_units,rand()%df_units->count);
+        list_clear(df_units);
+        list_add( df_units, entry );
+    }
 }
 
 /*
@@ -937,10 +1074,29 @@ int unit_check_merge( Unit *unit, Unit *source )
     if ( unit->prop.mov_type != source->prop.mov_type ) return 0;
     /* the unit strength must not exceed limit */
     if ( unit->str + source->str > 13 ) return 0;
+    /* fortresses (unit-class 7) could not merge */
+    if ( unit->prop.class == 7 ) return 0;
+    /* artillery with different ranges may not merge */
+    if (unit->prop.flags&ARTILLERY && unit->prop.rng!=source->prop.rng) return 0;
     /* not failed so far: allow merge */
     return 1;
 }
 
+/*
+====================================================================
+Get the maximum strength the unit can give for a split in its
+current state. Unit must have at least strength 3 remaining.
+====================================================================
+*/
+int unit_get_split_strength( Unit *unit )
+{
+    if ( unit->embark != EMBARK_NONE ) return 0;
+    if ( !unit->unused ) return 0;
+    if ( unit->str <= 4 ) return 0;
+    if ( unit->prop.class == 7 ) return 0; /* fortress */
+    return unit->str - 4;
+}
+
 /*
 ====================================================================
 Merge these two units: unit is the new unit and source must be
@@ -959,10 +1115,15 @@ void unit_merge( Unit *unit, Unit *source )
     weight1 /= total; weight2 /= total;
     /* no other actions allowed */
     unit->unused = 0; unit->cur_mov = 0; unit->cur_atk_count = 0;
+    /* update cost since used for gaining prestige */
+    unit->prop.cost = (unit->prop.cost * unit->str + 
+					source->prop.cost * source->str) / 
+					(unit->str + source->str);
     /* repair damage */
     unit->str += source->str;
-    /* reorganization costs all entrenchment */
-    unit->entr = 0;
+    /* reorganization costs some entrenchment: the new units are assumed to have
+       entrenchment 0 since they come. new entr is rounded weighted sum */
+    unit->entr = floor((float)unit->entr*weight1+0.5); /* + 0 * weight2 */
     /* update experience */
     i = (int)( weight1 * unit->exp + weight2 * source->exp );
     unit->exp = 0; unit_add_exp( unit, i );
@@ -1054,6 +1215,8 @@ int unit_check_rugged_def( Unit *unit, Unit *target )
         return 0;
     if ( ( unit->sel_prop->flags & SWIMMING ) || ( target->sel_prop->flags & SWIMMING ) )
         return 0;
+    if (unit->sel_prop->flags & ARTILLERY ) return 0; /* no rugged def against range attack */
+    if (unit->sel_prop->flags&IGNORE_ENTR) return 0; /* no rugged def for pioneers and such */
     if ( !unit_is_close( unit, target ) ) return 0;
     if ( target->entr == 0 ) return 0;
     return 1;
@@ -1068,9 +1231,114 @@ int unit_get_rugged_def_chance( Unit *unit, Unit *target )
 {
     /* PG's formula is
        5% * def_entr * 
-       ( (def_exp_level + 2) / (atk_exp_level + 2) )
+       ( (def_exp_level + 2) / (atk_exp_level + 2) ) *
        ( (def_entr_rate + 1) / (atk_entr_rate + 1) ) */
     return (int)( 5.0 * target->entr *
            ( (float)(target->exp_level + 2) / (unit->exp_level + 2) ) *
            ( (float)(target->sel_prop->entr_rate + 1) / (unit->sel_prop->entr_rate + 1) ) );
 }
+
+/*
+====================================================================
+Calculate the used fuel quantity. 'cost' is the base fuel cost to be
+deducted by terrain movement. The cost will be adjusted as needed.
+====================================================================
+*/
+int unit_calc_fuel_usage( Unit *unit, int cost )
+{
+    int used = cost;
+    
+    /* air units use up *at least* the half of their initial movement points.
+     */
+    if ( unit->sel_prop->flags & FLYING ) {
+        int half = unit->sel_prop->mov / 2;
+        if ( used < half ) used = half;
+    }
+    
+    /* ground units face a penalty during bad weather */
+    if ( !(unit->sel_prop->flags & SWIMMING)
+         && !(unit->sel_prop->flags & FLYING)
+         && weather_types[scen_get_weather()].flags & DOUBLE_FUEL_COST )
+        used *= 2;
+    return used;
+}
+
+/*
+====================================================================
+Update unit bar.
+====================================================================
+*/
+void unit_update_bar( Unit *unit )
+{
+    update_bar(unit);
+}
+
+/*
+====================================================================
+Disable all actions.
+====================================================================
+*/
+void unit_set_as_used( Unit *unit )
+{
+    unit->unused = 0; 
+    unit->cur_mov = 0; 
+    unit->cur_atk_count = 0;
+}
+
+/*
+====================================================================
+Duplicate the unit.
+====================================================================
+*/
+Unit *unit_duplicate( Unit *unit )
+{
+    Unit *new = calloc(1,sizeof(Unit));
+    memcpy(new,unit,sizeof(Unit));
+    unit_set_generic_name(new, units->count + 1, unit->prop.name);
+    if (unit->sel_prop==&unit->prop)
+        new->sel_prop=&new->prop;
+    else
+        new->sel_prop=&new->trsp_prop;
+    new->backup = calloc( 1, sizeof( Unit ) );
+    /* terrain can't be updated here */
+    return new;
+}
+
+/*
+====================================================================
+Check if unit has low ammo or fuel.
+====================================================================
+*/
+int unit_low_fuel( Unit *unit )
+{
+    if ( !unit_check_fuel_usage( unit ) )
+        return 0;
+    if ( unit->sel_prop->flags & FLYING ) {
+        if ( unit->cur_fuel <= 20 )
+            return 1;
+        return 0;
+    }
+    if ( unit->cur_fuel <= 10 )
+        return 1;
+    return 0;
+}
+int unit_low_ammo( Unit *unit )
+{
+    /* a unit is low on ammo if it has less than twenty percent of its
+     * class' ammo supply left, or less than two quantities,
+     * whatever value is lower
+     */
+    int percentage = unit->sel_prop->ammo / 5;
+    return unit->embark == EMBARK_NONE && unit->cur_ammo <= MINIMUM( percentage, 2 );
+}
+
+/*
+====================================================================
+Check whether unit can be considered for deployment.
+====================================================================
+*/
+int unit_supports_deploy( Unit *unit )
+{
+    return !(unit->prop.flags & SWIMMING) /* ships and */
+           && unit->prop.mov > 0; /* fortresses cannot be deployed */
+}
