@@ -18,6 +18,8 @@
 #include "lgeneral.h"
 #include "parser.h"
 #include "unit_lib.h"
+#include "localize.h"
+#include "nation.h"
 
 /*
 ====================================================================
@@ -85,6 +87,9 @@ StrToFlag fct_units[] = {
     { "tank", TANK },
     { "anti_tank", ANTI_TANK },
     { "suppr_fire", SUPPR_FIRE },
+    { "turn_suppr", TURN_SUPPR },
+    { "jet", JET },
+    { "ground_trsp_ok", GROUND_TRSP_OK },
     { "X", 0}
 };
 
@@ -155,10 +160,10 @@ static void unit_lib_delete_entry( void *ptr )
 Evaluate the unit. This score will become a relative one whereas
 the best rating will be 1000. Each operational region 
 ground/sea/air will have its own reference.
-This evaluation is PG specific.
+This evaluation is PG specific. Cost is not considered.
 ====================================================================
 */
-static void unit_lib_eval_unit( Unit_Lib_Entry *unit )
+void unit_lib_eval_unit( Unit_Lib_Entry *unit )
 {
     int attack = 0, defense = 0, misc = 0;
     /* The score is computed by dividing the unit's properties
@@ -221,7 +226,7 @@ mov_types and unit classes will be loaded (may only happen once)
 int unit_lib_load( char *fname, int main )
 {
     int i, j, icon_id;
-    SDL_Surface *icons;
+    SDL_Surface *icons = NULL;
     int icon_type;
     int width, height, offset;
     Uint32 color_key;
@@ -235,6 +240,7 @@ int unit_lib_load( char *fname, int main )
     PData *pd, *sub, *subsub;
     char path[512];
     char *str, *flag;
+    char *domain = 0;
     float scale;
     /* log info */
     int  log_dot_limit = 40; /* maximum of dots */
@@ -244,31 +250,34 @@ int unit_lib_load( char *fname, int main )
     char log_str[128];
     /* there can be only one main library */
     if ( main == UNIT_LIB_MAIN && unit_lib_main_loaded ) {
-        fprintf( stderr, "%s: can't load as main unit library (which is already loaded): loading as 'additional' instead\n",
+        fprintf( stderr, tr("%s: can't load as main unit library (which is already loaded): loading as 'additional' instead\n"),
                  fname );
         main = UNIT_LIB_ADD;
     }
     /* parse file */
-    sprintf( path, "%s/units/%s", SRC_DIR, fname );
-    sprintf( log_str, "  Parsing '%s'", fname );
+    sprintf( path, "%s/units/%s", get_gamedir(), fname );
+    sprintf( log_str, tr("  Parsing '%s'"), fname );
     write_line( sdl.screen, log_font, log_str, log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( ( pd = parser_read_file( fname, path ) ) == 0 ) goto parser_failure;
+    domain = determine_domain(pd, fname);
+    locale_load_domain(domain, 0/*FIXME*/);
     /* if main read target types & co */
     if ( main == UNIT_LIB_MAIN ) {
-        write_line( sdl.screen, log_font, "  Loading Main Definitions", log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
+        write_line( sdl.screen, log_font, tr("  Loading Main Definitions"), log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
         unit_lib = list_create( LIST_AUTO_DELETE, unit_lib_delete_entry );
         /* target types */
         if ( !parser_get_entries( pd, "target_types", &entries ) ) goto parser_failure;
         trgt_type_count = entries->count;
         if ( trgt_type_count > TARGET_TYPE_LIMIT ) {
-            fprintf( stderr, "%i target types is the limit!\n", TARGET_TYPE_LIMIT );
+            fprintf( stderr, tr("%i target types is the limit!\n"), TARGET_TYPE_LIMIT );
             trgt_type_count = TARGET_TYPE_LIMIT;
         }
         trgt_types = calloc( trgt_type_count, sizeof( Trgt_Type ) );
         list_reset( entries ); i = 0;
         while ( ( sub = list_next( entries ) ) ) {
             trgt_types[i].id = strdup( sub->name );
-            if ( !parser_get_string( sub, "name", &trgt_types[i].name ) ) goto parser_failure;
+            if ( !parser_get_value( sub, "name", &str, 0 ) ) goto parser_failure;
+            trgt_types[i].name = strdup(trd(domain, str));
             i++;
         }
         /* movement types */
@@ -277,7 +286,8 @@ int unit_lib_load( char *fname, int main )
         list_reset( entries ); mov_type_count = 0;
         while ( ( sub = list_next( entries ) ) ) {
             mov_types[mov_type_count].id = strdup( sub->name );
-            if ( !parser_get_string( sub, "name", &mov_types[mov_type_count].name ) ) goto parser_failure;
+            if ( !parser_get_value( sub, "name", &str, 0 ) ) goto parser_failure;
+            mov_types[mov_type_count].name = strdup(trd(domain, str));
 #ifdef WITH_SOUND                
             if ( parser_get_value( sub, "sound", &str, 0 ) ) {
                 mov_types[mov_type_count].wav_move = wav_load( str, 0 );
@@ -291,7 +301,16 @@ int unit_lib_load( char *fname, int main )
         list_reset( entries ); unit_class_count = 0;
         while ( ( sub = list_next( entries ) ) ) {
             unit_classes[unit_class_count].id = strdup( sub->name );
-            if ( !parser_get_string( sub, "name", &unit_classes[unit_class_count++].name ) ) goto parser_failure;
+            if ( !parser_get_value( sub, "name", &str, 0 ) ) goto parser_failure;
+            unit_classes[unit_class_count].name = strdup(trd(domain, str));
+	    unit_classes[unit_class_count].purchase = UC_PT_NONE;
+	    if (parser_get_value( sub, "purchase", &str, 0 )) {
+		    if (strcmp(str,"trsp") == 0)
+			    unit_classes[unit_class_count].purchase = UC_PT_TRSP;
+		    else if (strcmp(str,"normal") == 0)
+			    unit_classes[unit_class_count].purchase = UC_PT_NORMAL;
+	    }
+	    unit_class_count++;
             /* ignore sounds so far */
         }
         /* unit map tile icons */
@@ -307,16 +326,21 @@ int unit_lib_load( char *fname, int main )
         if ( !parser_get_value( pd, "move_icon", &str, 0 ) ) goto parser_failure;
         sprintf( path, "units/%s", str );
         if ( ( unit_info_icons->mov = load_surf( path, SDL_SWSURFACE ) ) == 0 ) goto failure; 
+        if ( !parser_get_value( pd, "guard_icon", &str, 0 ) ) str = "pg_guard.bmp";
+        sprintf( path, "units/%s", str );
+        if ( ( unit_info_icons->guard = load_surf( path, SDL_SWSURFACE ) ) == 0 ) goto failure; 
     }
     /* icons */
     if ( !parser_get_value( pd, "icon_type", &str, 0 ) ) goto parser_failure;
-    if ( STRCMP( str, "single" ) )
+    if (STRCMP( str, "fixed" ) )
+        icon_type = UNIT_ICON_FIXED;
+    else if ( STRCMP( str, "single" ) )
         icon_type = UNIT_ICON_SINGLE;
     else
         icon_type = UNIT_ICON_ALL_DIRS;
     if ( !parser_get_value( pd, "icons", &str, 0 ) ) goto parser_failure;
     sprintf( path, "units/%s", str );
-    write_line( sdl.screen, log_font, "  Loading Tactical Icons", log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
+    write_line( sdl.screen, log_font, tr("  Loading Tactical Icons"), log_x, &log_y ); refresh_screen( 0, 0, 0, 0 );
     if ( ( icons = load_surf( path, SDL_SWSURFACE ) ) == 0 ) goto failure; 
     /* unit lib entries */
     if ( !parser_get_entries( pd, "unit_lib", &entries ) ) goto parser_failure;
@@ -332,7 +356,15 @@ int unit_lib_load( char *fname, int main )
         /* identification */
         unit->id = strdup( sub->name );
         /* name */
-        if ( !parser_get_string( sub, "name", &unit->name ) ) goto parser_failure;
+        if ( !parser_get_value( sub, "name", &str, 0) ) goto parser_failure;
+        unit->name = strdup(trd(domain, str));
+	/* nation (if not found or 'none' unit can't be purchased) */
+	unit->nation = -1; /* undefined */
+	if ( parser_get_value( sub, "nation", &str, 0) && strcmp(str,"none") ) {
+		Nation *n = nation_find( str );
+		if (n)
+			unit->nation = nation_get_index( n );
+	}
         /* class id */
         unit->class = 0;
         if ( !parser_get_value( sub, "class", &str, 0 ) ) goto parser_failure;
@@ -394,11 +426,14 @@ int unit_lib_load( char *fname, int main )
         else
             if ( unit->flags & INFANTRY )
                 unit->entr_rate = 3;
-        /* time period of usage */
-        unit->start_year = unit->start_month = unit->end_year = 0;
+        /* time period of usage (0 == cannot be purchased) */
+        unit->start_year = unit->start_month = unit->last_year = 0;
         parser_get_int( sub, "start_year", &unit->start_year );
         parser_get_int( sub, "start_month", &unit->start_month );
-        parser_get_int( sub, "end_year", &unit->end_year );
+        parser_get_int( sub, "last_year", &unit->last_year );
+	/* cost of unit (0 == cannot be purchased) */
+	unit->cost = 0;
+	parser_get_int( sub, "cost", &unit->cost );
         /* icon */
         /* icon id */
         if ( !parser_get_int( sub, "icon_id", &icon_id ) ) goto parser_failure;
@@ -476,6 +511,8 @@ int unit_lib_load( char *fname, int main )
         /* read sounds -- well as there are none so far ... */
 #ifdef WITH_SOUND
         if ( parser_get_value( sub, "move_sound", &str, 0 ) ) {
+            // FIXME reloading the same sound more than once is a
+            // big waste of loadtime, runtime, and memory
             if ( ( unit->wav_move = wav_load( str, 0 ) ) )
                 unit->wav_alloc = 1;
             else {
@@ -533,12 +570,16 @@ int unit_lib_load( char *fname, int main )
                 unit->eval_score /= best_ground;
         }
     }
+    free(domain);
+    SDL_FreeSurface(icons);
     return 1;
 parser_failure:        
     fprintf( stderr, "%s\n", parser_get_error() );
 failure:
     unit_lib_delete();
     if ( pd ) parser_free( &pd );
+    free(domain);
+    SDL_FreeSurface(icons);
     return 0;
 }
 
@@ -586,6 +627,7 @@ void unit_lib_delete( void )
         if ( unit_info_icons->str ) SDL_FreeSurface( unit_info_icons->str );
         if ( unit_info_icons->atk ) SDL_FreeSurface( unit_info_icons->atk );
         if ( unit_info_icons->mov ) SDL_FreeSurface( unit_info_icons->mov );
+        if ( unit_info_icons->guard ) SDL_FreeSurface( unit_info_icons->guard );
         free( unit_info_icons );
         unit_info_icons = 0;
     }
